""" Содержит разные глобальные переменные проекта. """

import logging

DEFAULT_PORT = 7777
""" Порт по умолчанию для сетевого взаимодействия """

DEFAULT_IP_ADDRESS = '127.0.0.1'
""" IP адрес по умолчанию для подключения клиента """

MAX_CONNECTIONS = 5
""" Максимальная очередь подключений. """

MAX_PACKAGE_LENGTH = 1024
""" Максимальная длинна сообщения в байтах """

ENCODING = 'utf-8'
""" Кодировка проекта """

LOGGING_LEVEL = logging.DEBUG
"""Текущий уровень логирования"""

SERVER_CONFIG = 'server.ini'
"""База данных для хранения данных сервера:"""


ACTION = 'action'
TIME = 'time'
USER = 'user'
ACCOUNT_NAME = 'account_name'
SENDER = 'from'
DESTINATION = 'to'
DATA = 'bin'
PUBLIC_KEY = 'pubkey'

PRESENCE = 'presence'
RESPONSE = 'response'
ERROR = 'error'
MESSAGE = 'message'
MESSAGE_TEXT = 'mess_text'
EXIT = 'exit'
GET_CONTACTS = 'get_contacts'
LIST_INFO = 'data_list'
REMOVE_CONTACT = 'remove'
ADD_CONTACT = 'add'
USERS_REQUEST = 'get_users'
PUBLIC_KEY_REQUEST = 'pubkey_need'

# Словари - ответы:
RESPONSE_200 = {RESPONSE: 200}
"""Ответ 200."""


RESPONSE_202 = {RESPONSE: 202,
                LIST_INFO: None
                }
"""Ответ 202."""

RESPONSE_400 = {
    RESPONSE: 400,
    ERROR: None
}
"""Ответ 400."""

RESPONSE_205 = {
    RESPONSE: 205
}
"""Ответ 205."""

RESPONSE_511 = {
    RESPONSE: 511,
    DATA: None
}
"""Ответ 511."""



from sqlalchemy import Column, Integer, String, ForeignKey, Date, DateTime, \
    create_engine, text
from sqlalchemy.orm import declarative_base, sessionmaker
import datetime
Base = declarative_base()


class ClientDatabase:
    """ Класс структуры данных для клиентского приложения"""

    def __init__(self, username):

        self.database_engine = \
            create_engine(f'sqlite:///client_{username}.db3',
                          echo=False, pool_recycle=7200,
                          connect_args={'check_same_thread': False})
        Base.metadata.create_all(self.database_engine)

        session = sessionmaker(bind=self.database_engine)
        self.session = session()

    class KnownUsers(Base):
        """ Класс - отображение таблицы пользователей системы """
        __tablename__ = 'known_users'

        id = Column(Integer, primary_key=True)
        username = Column(String(25), nullable=True, unique=True, index=True)
        nick = Column(String(25), nullable=True, unique=False)
        first_name = Column(String(25), nullable=True, unique=False)
        last_name = Column(String(25), nullable=True, unique=False)
        birthday = Column(Date, nullable=True, unique=False)

        def __init__(self, user):
            self.username = user

    class MessageHistory(Base):
        """ Класс - отображение таблицы истории сообщений """
        __tablename__ = 'message_history'

        id = Column(Integer, primary_key=True)
        from_user_id = Column(ForeignKey('known_users.id'), nullable=True)
        to_user_id = Column(ForeignKey('known_users.id'), nullable=True)
        message = Column(String(256))
        datetime = Column(DateTime, nullable=False, unique=False)

        def __init__(self, from_user_id, to_user_id, message):
            self.from_user_id = from_user_id
            self.to_user_id = to_user_id
            self.message = message
            self.datetime = datetime.datetime.now()

    class Contacts(Base):
        """ Класс - отображение таблицы контактов """
        __tablename__ = 'contacts'
        id = Column(Integer, primary_key=True)
        user_id = Column(ForeignKey('known_users.id'), nullable=True)

        def __init__(self, user_id):
            self.user_id = user_id

    def add_contact(self, contact):
        """ Функция добавления контактов """
        user_id = self.session.query(self.KnownUsers).\
            filter_by(username=contact).\
            first().id
        if id:
            contact_row = self.Contacts(user_id)
            self.session.add(contact_row)
            self.session.commit()

    def del_contact(self, contact):
        """ Функция удаления контакта """
        if contact:
            user_id = self.session.query(self.KnownUsers).\
                filter_by(username=contact).all()[0].id
            self.session.query(self.Contacts).\
                filter_by(user_id=user_id).delete()
            self.session.commit()

    def add_users(self, users_list):
        """ Функция добавления известных пользователей.
        Пользователи получаются только с сервера, поэтому таблица очищается.
        """
        self.session.query(self.KnownUsers).delete()
        for user in users_list:
            user_row = self.KnownUsers(user)
            self.session.add(user_row)
        self.session.commit()

    def save_message(self, from_user, to_user, message):
        """ Функция сохраняющая сообщения """
        from_user_id = self.session.query(self.KnownUsers).\
            filter_by(username=from_user).all()[0].id
        to_user_id = self.session.query(self.KnownUsers).\
            filter_by(username=to_user).all()[0].id
        message_row = self.MessageHistory(from_user_id=from_user_id,
                                          to_user_id=to_user_id,
                                          message=message)
        self.session.add(message_row)
        self.session.commit()

    def get_contacts(self):
        """ Функция возвращающая контакты """
        # TODO Проверить соединение INNER JOIN
        query = self.session.query(self.KnownUsers.username).\
            join(self.Contacts)
        query_result = query.all()
        return [contact[0] for contact in query_result]

    def get_users(self):
        """ Функция возвращающая список известных пользователей """
        return [user[0]
                for user in self.session.query(self.KnownUsers.username).all()]

    def check_user(self, user):
        """ Функция проверяющая наличие пользователя в известных """
        if self.session.query(self.KnownUsers).\
                filter_by(username=user).count():
            return True
        else:
            return False

    def check_contact(self, username):
        """" Функция проверяющая наличие пользователя в известных """
        if self.session.query(self.Contacts).join(self.KnownUsers).\
                filter(self.KnownUsers.username == username).count():
            return True
        else:
            return False

    def get_history(self, from_who=None, to_who=None):
        """ Функция возвращающая историю переписки """
        where_str = ''
        where_str += f"from_user.username = '{from_who}'" if from_who else ''
        where_str += ' or ' if from_who and to_who else ''
        where_str += f"to_user.username = '{to_who}'" if to_who else ''

        sql = text(f'''
        SELECT DISTINCT
            from_user.username AS from_user
            ,to_user.username AS to_user
            ,message_history.message AS message
            ,message_history.datetime AS datetime
        FROM
            message_history
            JOIN known_users AS from_user ON 
            message_history.from_user_id = from_user.id 
            JOIN known_users AS to_user ON 
            message_history.to_user_id = to_user.id
        WHERE {where_str}
        ''')
        query_result = self.session.execute(sql)

        return [(history_row.from_user,
                 history_row.to_user,
                 history_row.message,
                 datetime.datetime.strptime(history_row.datetime,
                                            '%Y-%m-%d %H:%M:%S.%f').
                 replace(microsecond=0))
                for history_row in query_result]
        # TODO разобраться с ORM, связи с дух таблиц


if __name__ == '__main__':
    test_db = ClientDatabase('test1')
    test_db.add_users(['test1', 'test2', 'test3', 'test4', 'test5'])
    for i in ['test3', 'test4', 'test5']:
        test_db.add_contact(i)
    test_db.add_contact('test4')
    test_db.\
        save_message('test1',
                     'test2',
                     f'Привет! я тестовое сообщение от '
                     f'{datetime.datetime.now()}!')
    test_db.save_message('test2',
                         'test1',
                         f'Привет! я другое тестовое сообщение от '
                         f'{datetime.datetime.now()}!')
    print(test_db.get_contacts())
    print(test_db.get_users())
    print(test_db.check_user('test1'))
    print(test_db.check_user('test10'))
    print(test_db.get_history('test2'))
    print(test_db.get_history(to_who='test2'))
    print(test_db.get_history('test3'))
    test_db.del_contact('test4')
    print(test_db.get_contacts())



from setuptools import setup, find_packages

setup(name="ship_messenger_client",
      version="0.8.7",
      description="ship messenger client",
      author="Vladimir Ruzhelovich",
      author_email="Ruzhelovich.Vladimir@gmail.com",
      packages=find_packages(),
      install_requires=['PyQt5', 'sqlalchemy', 'pycryptodome', 'pycryptodomex']
      )



import dis


class ClientVerifier(type):
    """ Мета-класс для верификации клиентского приложения  """

    def __init__(cls, cls_name, bases, cls_dict):
        methods = []
        attrs = []
        for func in cls_dict:
            # Почему-то падает на дос строке
            if '__doc__' == func:
                continue
            try:
                ret = dis.get_instructions(cls_dict[func])
            except TypeError:
                pass
            else:
                for i in ret:
                    # print(i)
                    if i.opname == 'LOAD_GLOBAL':
                        if i.argval not in methods:
                            methods.append(i.argval)
                    elif i.opname == 'LOAD_ATTR':
                        if i.argval not in attrs:
                            attrs.append(i.argval)
        # print(methods)
        if 'accept' in methods and 'listen' in methods and 'socket' in methods:
            raise TypeError(
                'Использование методов accept, listen, socket не допустимо на '
                'клиенте')
        # TODO перестало работать контроль инициализации сокета, разобраться
        # if not ('SOCK_STREAM' in attrs and 'AF_INET' in attrs):
        #     print("*"*20, attrs, "*"*20)
        #     raise TypeError('Не корректная инициализация сокета.')
        super().__init__(cls_name, bases, cls_dict)


class ServerVerifier(type):
    """ Мета-класс для верификации серверного приложения  """

    def __init__(cls, cls_name, bases, cls_dict):
        methods = []
        attrs = []
        for func in cls_dict:
            # Почему-то падает на дос строке
            if '__doc__' == func:
                continue
            try:
                ret = dis.get_instructions(cls_dict[func])
            except TypeError:
                pass
            else:
                for i in ret:
                    # print(i)
                    if i.opname == 'LOAD_GLOBAL':
                        if i.argval not in methods:
                            methods.append(i.argval)
                    elif i.opname == 'LOAD_ATTR':
                        if i.argval not in attrs:
                            attrs.append(i.argval)
        # print(methods)
        if 'connect' in methods:
            raise TypeError(
                'Использование метода connect не допустимо в серверном классе')
        if not ('SOCK_STREAM' in attrs and 'AF_INET' in attrs):
            raise TypeError('Не корректная инициализация сокета.')
        super().__init__(cls_name, bases, cls_dict)



from PyQt5.QtWidgets import QDialog, QPushButton, QLineEdit, QApplication, QLabel , qApp


class UserNameDialog(QDialog):
    def __init__(self):
        super().__init__()

        self.ok_pressed = False
        self.setWindowTitle('Messanger client (Beta)')
        self.setFixedSize(350, 160)

        self.label = QLabel('Enter username:', self)
        self.label.setFixedSize(330, 20)
        self.label.move(10, 10)

        self.client_name = QLineEdit(self)
        self.client_name.setFixedSize(330, 20)
        self.client_name.move(10, 35)

        self.label_passwd = QLabel('Enter password:', self)
        self.client_name.setFixedSize(330, 20)
        self.label_passwd.move(10, 60)

        self.client_passwd = QLineEdit(self)
        self.client_passwd.setFixedSize(330, 20)
        self.client_passwd.move(10, 85)
        self.client_passwd.setEchoMode(QLineEdit.Password)

        self.btn_ok = QPushButton('Continue..', self)
        self.btn_ok.move(170, 120)
        self.btn_ok.clicked.connect(self.click)

        self.btn_cancel = QPushButton('Exit', self)
        self.btn_cancel.move(260, 120)
        self.btn_cancel.clicked.connect(qApp.exit)
        self.show()

    def click(self):
        if self.client_name.text():
            self.ok_pressed = True
            qApp.exit()


if __name__ == '__main__':
    app = QApplication([])
    dial = UserNameDialog()
    app.exec_()



"""
Модуль диалога добавления контактов
"""

import sys
import logging

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QDialog, QLabel, QComboBox, QPushButton

sys.path.append('../')

logger = logging.getLogger('client')


"""Диалог выбора контакта для добавления"""


class AddContactDialog(QDialog):
    """Диалог выбора контакта для добавления"""

    def __init__(self, transport, database):
        super().__init__()
        self.transport = transport
        self.database = database

        self.setFixedSize(350, 120)
        self.setWindowTitle('Выберите контакт для добавления:')
        self.setAttribute(Qt.WA_DeleteOnClose)
        self.setModal(True)

        self.selector_label = QLabel('Выберите контакт для добавления:', self)
        self.selector_label.setFixedSize(200, 20)
        self.selector_label.move(10, 0)

        self.selector = QComboBox(self)
        self.selector.setFixedSize(200, 20)
        self.selector.move(10, 30)

        self.btn_refresh = QPushButton('Обновить список', self)
        self.btn_refresh.setFixedSize(100, 30)
        self.btn_refresh.move(60, 60)

        self.btn_ok = QPushButton('Добавить', self)
        self.btn_ok.setFixedSize(100, 30)
        self.btn_ok.move(230, 20)

        self.btn_cancel = QPushButton('Отмена', self)
        self.btn_cancel.setFixedSize(100, 30)
        self.btn_cancel.move(230, 60)
        self.btn_cancel.clicked.connect(self.close_windows)

        # Заполняем список возможных контактов
        self.possible_contacts_update()
        # Назначаем действие на кнопку обновить
        self.btn_refresh.clicked.connect(self.update_possible_contacts)

    def close_windows(self):
        super().close()

    def possible_contacts_update(self):
        """Заполняем список возможных контактов"""
        self.selector.clear()
        # множества всех контактов и контактов клиента
        contacts_list = set(self.database.get_contacts())
        users_list = set(self.database.get_users())
        # Удалим сами себя из списка пользователей
        if self.transport.username in users_list:
            users_list.remove(self.transport.username)
        # Добавляем список возможных контактов
        self.selector.addItems(users_list - contacts_list)

    # Обновляет таблицу известных пользователей
    def update_possible_contacts(self):
        """Обновляет таблицу известных пользователей"""
        try:
            self.transport.user_list_update()
        except OSError:
            pass
        else:
            logger.debug('Обновление списка пользователей с сервера выполнено')
            self.possible_contacts_update()



"""
Реализовать дескриптор для класса серверного сокета, а в нем — проверку номера
порта. Это должно быть целое число (>=0). Значение порта по умолчанию равняется
7777. Дескриптор надо создать в отдельном классе. Его экземпляр добавить в
пределах класса серверного сокета. Номер порта передается в экземпляр
дескриптора при запуске сервера.
"""
from common.host_ping import is_available_host


class Port:
    """ Дескриптор контроля порта """

    def __set__(self, instance, value):
        if not 1023 < value < 65536:
            raise ValueError("Допустимы значение порта с 1024 до 65535")
        instance.__dict__[self.name] = value

    def __set_name__(self, owner, name):
        self.name = name


class Host:
    """ Дескриптор контроля хоста """

    def __set__(self, instance, value):
        if not is_available_host(value):
            raise ValueError("Хост не доступен")
        instance.__dict__[self.name] = value

    def __set_name__(self, owner, name):
        self.name = name



"""
Модуль проверок доступности хоста
"""
from ipaddress import ip_address
from subprocess import Popen, PIPE, SubprocessError
import json
from tabulate import tabulate


def host_range_ping(start_ip, count):
    """
    Функция для перебора ip-адресов из заданного диапазона.
    По результатам
    проверки должно выводиться соответствующее сообщение.
    """
    result = []

    try:
        _ip_address = ip_address(start_ip)
    except ValueError:
        return []

    for num in range(count):
        result.append(str(_ip_address+num))

    return host_ping(result)


def is_available_host(host, timeout=3):
    """
    Проверка доступности хоста
    """
    try:
        process = Popen(('ping', f'{host}', '-c1',
                         f'-W{timeout}'), shell=False, stdout=PIPE)
        process.wait()
    except SubprocessError:
        return False

    return process.returncode == 0


def host_ping(host_list=None, timeout=3):
    """
    Функцию host_ping(), в которой с помощью утилиты ping будет проверяться
    доступность сетевых узлов. Аргументом функции является список, в котором
    каждый сетевой узел должен быть представлен именем хоста или ip-адресом
    """
    if host_list is None:
        host_list = []
    result = {}
    for host in host_list:

        status = "Reachable" if is_available_host(host, timeout) \
            else 'Unreachable'
        if status in result:
            result[status].append(host)
        else:
            result[status] = [host]
    return result


def host_range_ping_tab(start_ip, count):
    """
    Возвращает таблицу доступности адресов
    """
    return tabulate(host_range_ping(start_ip, count), headers='keys')


if __name__ == '__main__':

    res = host_ping(['localhost'], 1)
    print(json.dumps(res, indent=4))



from common.errors import ServerError
from common.utils import *
import binascii
import hashlib
import hmac
import socket
import sys
import time
import threading
from PyQt5.QtCore import pyqtSignal, QObject

sys.path.append('../')

logger = logging.getLogger('client')
socket_lock = threading.Lock()


class ClientTransport(threading.Thread, QObject):
    """ Class of interaction between client app and server app """

    # Signal of new client message
    new_message = pyqtSignal(str)
    # Signal of lost client connection
    connection_lost = pyqtSignal()
    # Signal of 205 Reset Content
    message_205 = pyqtSignal()

    def __init__(self, port, address, database, username, passwd, keys):

        threading.Thread.__init__(self)
        QObject.__init__(self)

        self.server_address = address
        self.server_port = port
        self.database = database
        self.username = username
        self.password = passwd
        self.transport = None
        self.keys = keys

        # Запускаем процедуру авторизации
        # Получаем хэш пароля
        self.passwd_bytes = self.password.encode('utf-8')
        salt = self.username.lower().encode('utf-8')
        self.passwd_hash = hashlib.pbkdf2_hmac(
            'sha512', self.passwd_bytes, salt, 10000)
        self.passwd_hash_string = binascii.hexlify(self.passwd_hash)

        logger.debug(f'Passwd hash ready: {self.passwd_hash_string}')

        # Получаем публичный ключ и декодируем его из байтов
        self.pubkey = self.keys.publickey().export_key().decode('ascii')

        self.__connection_init()
        self.update_users_and_contact_list()
        self.running = True

    def __connection_init(self):
        """ Initialize the socket (transport)
        and notify the server about it """
        if not self.__init_transport():
            self.__critical_error('Failed to connect to server')
        logger.debug('The connection to the server is established.')
        try:
            with socket_lock:
                send_message(self.transport,
                             self.__create_presence())  # Send msg presence
                ans = get_message(self.transport)
                logger.debug(f'Server response = {ans}.')

                if RESPONSE in ans:
                    if ans[RESPONSE] == 400:
                        raise ServerError(ans[ERROR])
                    elif ans[RESPONSE] == 511:
                        # If 511 Network Authentication Required then
                        ans_data = ans[DATA]
                        digest = hmac.new(
                            self.passwd_hash_string,
                            ans_data.encode('utf-8'),
                            'MD5').digest()
                        my_ans = RESPONSE_511
                        my_ans[DATA] = binascii.b2a_base64(
                            digest).decode('ascii')
                        send_message(self.transport, my_ans)
                        ans = get_message(self.transport)
                        self.process_server_ans(ans)
        except (OSError, json.JSONDecodeError):
            self.__critical_error('Lost connection to server')
        logger.info('The connection to the server is established')

    def __init_transport(self):
        """ Initialize the socket (transport)"""
        self.transport = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.transport.settimeout(5)
        # return flag
        connected = False
        for i in range(5):
            logger.info(f'Attempt #{i + 1} to connect to server')
            try:
                self.transport.connect((self.server_address, self.server_port))
            except (OSError, ConnectionRefusedError):
                time.sleep(1)
                continue
            connected = True
            break

        # Если соединится не удалось - исключение
        if not connected:
            logger.critical('Не удалось установить соединение с сервером')
            raise ServerError('Не удалось установить соединение с сервером')

        logger.debug('Starting auth dialog.')

        return connected

    def __create_presence(self):
        """ Запрос о присутствии клиента """
        out = {
            ACTION: PRESENCE,
            TIME: time.time(),
            USER: {
                ACCOUNT_NAME: self.username,
                PUBLIC_KEY: self.pubkey
            }
        }
        logger.debug(
            f'{self.username}: Generated {PRESENCE} '
            f'message for user {self.username}')
        return out

    def process_server_ans(self, message):
        """ Функция обрабатывающая сообщения от сервера.
        Ничего не возвращает. Генерирует исключение при ошибке. """

        logger.debug(f'Parse messages from the server: {message}')

        if RESPONSE in message:
            if message[RESPONSE] == 200:
                return
            elif message[RESPONSE] == 400:
                raise ServerError(f'{message[ERROR]}')
            elif message[RESPONSE] == 205:
                self.user_list_update()
                self.contacts_list_update()
                self.message_205.emit()
            else:
                logger.debug(
                    f'Unknown confirmation code accepted: {message[RESPONSE]}')
        # Если это сообщение от пользователя, то добавляем в базу, даём сигнал
        # о новом сообщении
        elif ACTION in message and message[ACTION] == MESSAGE \
                and SENDER in message and DESTINATION in message \
                and MESSAGE_TEXT in message \
                and message[DESTINATION] == self.username:
            logger.debug(
                f'Received message from user: {message[SENDER]}, '
                f'"{message[MESSAGE_TEXT]}"')
            self.database.save_message(
                message[SENDER],
                self.username,
                message[MESSAGE_TEXT])
            self.new_message.emit(message[SENDER])

    def update_users_and_contact_list(self):
        """ Updating user and contact lists """
        try:
            self.user_list_update()
            self.contacts_list_update()
        except OSError as err:
            if err.errno:
                self.__critical_error('Lost connection to server')
            logger.error('Timeout when updating user and contact lists')
        except json.JSONDecodeError:
            self.__critical_error('Lost connection to server')

    def contacts_list_update(self):
        """ Updating contact lists """
        logger.debug(f'The request of contact list of user: {self.username}')
        req = {
            ACTION: GET_CONTACTS,
            TIME: time.time(),
            USER: self.username
        }
        logger.debug(f'Request generated: {req}')
        with socket_lock:
            send_message(self.transport, req)
            ans = get_message(self.transport)
        logger.debug(f'Received a response: {ans}')
        if RESPONSE in ans and ans[RESPONSE] == 202:
            for contact in ans[LIST_INFO]:
                self.database.add_contact(contact)
        else:
            logger.error(f'Failed to update contact list ({self.username})')

    def user_list_update(self):
        """ Updating the table of known users """
        logger.debug(f'The request of known users list ({self.username})')
        req = {
            ACTION: USERS_REQUEST,
            TIME: time.time(),
            ACCOUNT_NAME: self.username
        }
        with socket_lock:
            send_message(self.transport, req)
            ans = get_message(self.transport)
        if RESPONSE in ans and ans[RESPONSE] == 202:
            self.database.add_users(ans[LIST_INFO])
        else:
            logger.error(
                f'Failed to update known users list ({self.username})')

    def key_request(self, user):
        """ Метод запрашивающий с сервера публичный ключ пользователя. """
        logger.debug(f'Запрос публичного ключа для {user}')
        req = {
            ACTION: PUBLIC_KEY_REQUEST,
            TIME: time.time(),
            ACCOUNT_NAME: user
        }
        with socket_lock:
            send_message(self.transport, req)
            ans = get_message(self.transport)
        if RESPONSE in ans and ans[RESPONSE] == 511:
            return ans[DATA]
        else:
            logger.error(f'Не удалось получить ключ собеседника{user}.')

    def add_contact(self, contact):
        """ Добавление нового контакта """
        req = {
            ACTION: ADD_CONTACT,
            TIME: time.time(),
            USER: self.username,
            ACCOUNT_NAME: contact
        }
        logger.debug(f'Create contact {contact} ({self.username})')
        with socket_lock:
            send_message(self.transport, req)
            self.process_server_ans(get_message(self.transport))

    def remove_contact(self, contact):
        """ Удаление контакта """
        logger.debug(f'Delete contact {contact} ({self.username})')
        req = {
            ACTION: REMOVE_CONTACT,
            TIME: time.time(),
            USER: self.username,
            ACCOUNT_NAME: contact
        }
        with socket_lock:
            send_message(self.transport, req)
            self.process_server_ans(get_message(self.transport))

    def transport_shutdown(self):
        """ Отправка сообщения на сервер
        об закрытия соединения и выходе клиента"""
        self.running = False
        message = {
            ACTION: EXIT,
            TIME: time.time(),
            ACCOUNT_NAME: self.username
        }
        with socket_lock:
            try:
                send_message(self.transport, message)
            except OSError:
                pass
        logger.debug('The transport shutdown')
        time.sleep(0.5)

    def send_message(self, to, message):
        """ Sending a message to the server """
        message_dict = {
            ACTION: MESSAGE,
            SENDER: self.username,
            DESTINATION: to,
            TIME: time.time(),
            MESSAGE_TEXT: message
        }
        with socket_lock:
            send_message(self.transport, message_dict)
            self.process_server_ans(get_message(self.transport))
            logger.info(f'Message sent to user {to}')

    def run(self):
        logger.debug(
            'The process of receiving messages from the server has started...')
        while self.running:
            # Отдыхаем секунду и снова пробуем захватить сокет.
            # Если не сделать тут задержку, то отправка может достаточно долго
            # ждать освобождения сокета.
            time.sleep(1)
            with socket_lock:
                try:
                    self.transport.settimeout(0.5)
                    message = get_message(self.transport)
                except OSError as err:
                    if err.errno:
                        logger.critical(f'Lost connection to server.')
                        self.running = False
                        self.connection_lost.emit()
                except (ConnectionError, ConnectionAbortedError,
                        ConnectionResetError, json.JSONDecodeError, TypeError):
                    logger.critical(f'Lost connection to server.')
                    self.running = False
                    self.connection_lost.emit()
                else:
                    logger.debug(
                        f'Received a message from the server: {message}')
                    self.process_server_ans(message)
                finally:
                    self.transport.settimeout(5)

    @staticmethod
    def __critical_error(msg):
        """ Handling critical errors """
        logger.critical(msg)
        raise ServerError(msg)



import argparse
import os
import sys

from Crypto.PublicKey import RSA
from PyQt5.QtWidgets import QApplication

from common.variables import *
from common.errors import ServerError
from common.decos import log
from client.database import ClientDatabase
from client.transport import ClientTransport
from client.main_window import ClientMainWindow
from client.start_dialog import UserNameDialog

logger = logging.getLogger('client')


@log
def arg_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('address', default=DEFAULT_IP_ADDRESS, nargs='?')
    parser.add_argument('port', default=DEFAULT_PORT, type=int, nargs='?')
    parser.add_argument('-n', '--name', default=None, nargs='?')
    parser.add_argument('-p', '--password', default='', nargs='?')
    namespace = parser.parse_args(sys.argv[1:])
    if not 1023 < namespace.port < 65536:
        logger.critical(
            f'Invalid port {namespace.port}. '
            f'The available value is from 1024 to 65535.')
        exit(1)
    result = (namespace.address, namespace.port, namespace.name,
              namespace.password)
    return result


def get_user_security(username):

    # Get security key
    # Загружаем ключи с файла, если же файла нет, то генерируем новую пару.
    #dir_path = os.path.dirname(os.path.realpath(__file__))
    dir_path = os.getcwd()
    key_file = os.path.join(
        dir_path,
        'client',
        '.secrets',
        f'{username}.key')
    if not os.path.exists(key_file):
        user_keys = RSA.generate(2048, os.urandom)
        with open(key_file, 'wb') as key:
            key.write(user_keys.export_key())
    else:
        with open(key_file, 'rb') as key:
            user_keys = RSA.import_key(key.read())

    logger.debug("Keys successfully loaded.")
    return user_keys


if __name__ == '__main__':

    # TODO Решить баг с дублирование записей в списке контактов
    server_address, server_port, client_name, client_passwd = arg_parser()
    client_app = QApplication(sys.argv)

    if not client_name or not client_passwd:
        start_dialog = UserNameDialog()
        client_app.exec_()

        if start_dialog.ok_pressed:
            client_name = start_dialog.client_name.text()
            client_passwd = start_dialog.client_passwd.text()
            logger.debug(
                f'Using USERNAME = {client_name}, PASSWD = {client_passwd}.')
            del start_dialog
        else:
            exit(0)

    logger.info(
        f'Client application started with parameters server: '
        f'{server_address}:{server_port}, user name: {client_name}')

    keys = get_user_security(client_name)

    database = ClientDatabase(client_name)
    transport = None

    # Start demon
    try:
        transport = ClientTransport(
            server_port,
            server_address,
            database,
            client_name,
            client_passwd,
            keys)
    except ServerError as error:
        print(error.text)
        exit(1)
    transport.setDaemon(True)
    transport.start()

    # Start main application
    main_window = ClientMainWindow(database, transport)
    main_window.make_connection(transport)
    main_window.setWindowTitle(
        f'Messanger application (alpha release) - {client_name}')
    client_app.exec_()

    transport.transport_shutdown()
    transport.join()



"""
Другие утилиты
"""

# from common.decos import log
from common.errors import IncorrectDataReceivedError, NonDictInputError
from common.variables import *
import json
import sys
sys.path.append('../')

logger = logging.getLogger('utils')

# @log


def get_message(client):
    """
    Функция принимает сообщения в формате JSON от удалённых компьютеров,
    декодирует полученное сообщение и проверяет что получен словарь.
    """

    encoded_response = client.recv(MAX_PACKAGE_LENGTH)
    if not encoded_response:
        return {}
    if isinstance(encoded_response, bytes):
        json_response = encoded_response.decode(ENCODING)
        response = None
        try:
            response = json.loads(json_response)
        except Exception as err:
            logger.error(f'Не возможно считать json объект:\n{json_response} '
                         f'- {err}')
        if isinstance(response, dict):
            return response
        else:
            raise IncorrectDataReceivedError
    else:
        raise IncorrectDataReceivedError


# @log
def send_message(sock, message):
    """
    Функция отправки словарей через сокет. Кодирует словарь в формат JSON
    и отправляет через сокет.
    """

    if not isinstance(message, dict):
        raise NonDictInputError
    js_message = json.dumps(message)
    encoded_message = js_message.encode(ENCODING)
    sock.send(encoded_message)



import sys
import logging

# метод определения модуля, источника запуска.
if sys.argv[0].find('client') == -1:
    # если сервер
    logger = logging.getLogger('server')
else:
    # если клиент
    logger = logging.getLogger('client')


def log(func_to_log):
    def log_saver(*args, **kwargs):
        logger.debug(
            f'Была вызвана функция {func_to_log.__name__} '
            f'c параметрами {args}, '
            f'{kwargs}. Вызов из модуля {func_to_log.__module__}')
        ret = func_to_log(*args, **kwargs)
        return ret

    return log_saver



import sys
import json
import logging

from common.errors import ServerError
from client.del_contact import DelContactDialog
from client.add_contact import AddContactDialog
from client.main_window_conv import UiMainClientWindow
from Crypto.Cipher import PKCS1_OAEP
from Crypto.PublicKey import RSA
from PyQt5.QtWidgets import QMainWindow, qApp, QMessageBox
from PyQt5.QtGui import QStandardItemModel, QStandardItem, QBrush, QColor
from PyQt5.QtCore import pyqtSlot, Qt

sys.path.append('../')
logger = logging.getLogger('client')

global select_dialog
global remove_dialog

# Класс основного окна


class ClientMainWindow(QMainWindow):
    """ Класс основного окна """

    def __init__(self, database, transport):
        super().__init__()
        # основные переменные
        self.database = database
        self.transport = transport

        # Загружаем конфигурацию окна из дизайнера
        self.ui = UiMainClientWindow()
        self.ui.setup_ui(self)

        # Кнопка "Выход"
        self.ui.menu_exit.triggered.connect(qApp.exit)

        # Кнопка отправить сообщение
        self.ui.btn_send.clicked.connect(self.send_message)

        # "добавить контакт"
        self.ui.btn_add_contact.clicked.connect(self.add_contact_window)
        self.ui.menu_add_contact.triggered.connect(self.add_contact_window)

        # Удалить контакт
        self.ui.btn_remove_contact.clicked.connect(self.delete_contact_window)
        self.ui.menu_del_contact.triggered.connect(self.delete_contact_window)

        # Дополнительные требующиеся атрибуты
        self.contacts_model = None
        self.history_model = None
        self.messages = QMessageBox()

        self.current_chat = None
        self.current_chat_key = None
        self.encryptor = None

        self.ui.list_messages.setHorizontalScrollBarPolicy(
            Qt.ScrollBarAlwaysOff)
        self.ui.list_messages.setWordWrap(True)

        # Двойной клик по листу контактов отправляется в обработчик
        self.ui.list_contacts.clicked.connect(self.select_active_user)

        self.clients_list_update()
        self.set_disabled_input()
        self.show()

    def set_disabled_input(self):
        """ Деактивировать поля ввода """

        # Надпись  - получатель.
        self.ui.label_new_message.setText(
            'Для выбора получателя дважды кликните на нем в окне контактов.')
        self.ui.text_message.clear()
        if self.history_model:
            self.history_model.clear()

        # Поле ввода и кнопка отправки неактивны до выбора получателя.
        self.ui.btn_clear.setDisabled(True)
        self.ui.btn_send.setDisabled(True)
        self.ui.text_message.setDisabled(True)

    def history_list_update(self):
        """ Заполняем историю сообщений """
        # Получаем историю сортированную по дате
        history = self.database.get_history(from_who=self.current_chat,
                                            to_who=self.current_chat)
        history_list = sorted(history, key=lambda elem: elem[3])
        # Если модель не создана, создадим.
        if not self.history_model:
            self.history_model = QStandardItemModel()
            self.ui.list_messages.setModel(self.history_model)
        # Очистим от старых записей
        self.history_model.clear()
        # Берём не более 20 последних записей.
        length = len(history_list)
        start_index = 0
        if length > 20:
            start_index = length - 20
        # Заполнение модели записями, так-же стоит разделить входящие и
        # исходящие выравниванием и разным фоном.
        # Записи в обратном порядке, поэтому выбираем их с конца и не более 20
        for i in range(start_index, length):
            item = history_list[i]
            if item[1] != self.current_chat:
                mess = QStandardItem(f'Входящее от {item[3]}:\n {item[2]}')
                mess.setEditable(False)
                mess.setBackground(QBrush(QColor(255, 213, 213)))
                mess.setTextAlignment(Qt.AlignLeft)
                self.history_model.appendRow(mess)
            else:
                mess = QStandardItem(f'Исходящее от {item[3]}:\n {item[2]}')
                mess.setEditable(False)
                mess.setTextAlignment(Qt.AlignRight)
                mess.setBackground(QBrush(QColor(204, 255, 204)))
                self.history_model.appendRow(mess)
        self.ui.list_messages.scrollToBottom()


    def select_active_user(self):
        """ Функция активации пользователя """

        # Выбранный пользователем (двойной клик) находится в выделенном
        # элементе в QListView
        self.current_chat = self.ui.list_contacts.currentIndex().data()
        # вызываем основную функцию
        self.set_active_user()

    def set_active_user(self):
        """Функция устанавливающая активного собеседника"""
        # Запрашиваем публичный ключ пользователя и создаём объект шифрования
        self.current_chat_key = None
        self.encryptor = None
        try:
            self.current_chat_key = self.transport.key_request(
                self.current_chat)
            logger.debug(f'Загружен открытый ключ для {self.current_chat}')
            if self.current_chat_key:
                self.encryptor = PKCS1_OAEP.new(
                    RSA.import_key(self.current_chat_key))
        except (OSError, json.JSONDecodeError):
            logger.debug(f'Не удалось получить ключ для {self.current_chat}')

        # Если ключа нет, то ошибка, что не удалось начать чат с пользователем
        if not self.current_chat_key:
            self.messages.warning(
                self, 'Ошибка', 'Для выбранного пользователя '
                                'нет ключа шифрования.')
            return

        # Ставим надпись и активируем кнопки
        self.ui.label_new_message.setText(
            f'Введите сообщение для {self.current_chat}:')
        self.ui.btn_clear.setDisabled(False)
        self.ui.btn_send.setDisabled(False)
        self.ui.text_message.setDisabled(False)

        # Заполняем окно историю сообщений по требуемому пользователю.
        self.history_list_update()

    def clients_list_update(self):
        """ Функция обновляющая контакт лист """
        contacts_list = self.database.get_contacts()
        self.contacts_model = QStandardItemModel()
        for i in sorted(contacts_list):
            item = QStandardItem(i)
            item.setEditable(False)
            self.contacts_model.appendRow(item)
        self.ui.list_contacts.setModel(self.contacts_model)


    def add_contact_window(self):
        """ Функция добавления контакта """
        global select_dialog
        select_dialog = AddContactDialog(self.transport, self.database)
        select_dialog.btn_ok.clicked.connect(
            lambda: self.add_contact_action(select_dialog))
        select_dialog.show()

    def add_contact_action(self, item):
        """ Функция - обработчик добавления, сообщает серверу, обновляет
        таблицу и список контактов"""
        new_contact = item.selector.currentText()
        self.add_contact(new_contact)
        item.close()

    def add_contact(self, new_contact):
        """ Функция добавляющая контакт в базы """
        try:
            self.transport.add_contact(new_contact)
        except ServerError as err:
            self.messages.critical(self, 'Ошибка сервера', err.text)
        except OSError as err:
            if err.errno:
                self.messages.critical(
                    self, 'Ошибка', 'Потеряно соединение с сервером!')
                self.close()
            self.messages.critical(self, 'Ошибка', 'Таймаут соединения!')
        else:
            self.database.add_contact(new_contact)
            new_contact = QStandardItem(new_contact)
            new_contact.setEditable(False)
            self.contacts_model.appendRow(new_contact)
            logger.info(f'Успешно добавлен контакт {new_contact}')

    def delete_contact_window(self):
        """ Функция удаления контакта """
        global remove_dialog

        if self.current_chat:
            self.delete_contact(self.current_chat)
        else:
            remove_dialog = DelContactDialog(self.database)
            remove_dialog.btn_ok.clicked.connect(
                lambda: self.delete_contact(remove_dialog))
            remove_dialog.show()


    def delete_contact(self, item=None):
        """ Функция обработчик удаления контакта, сообщает на сервер,
        обновляет таблицу контактов"""

        # Если задан чат, то удаляем его, в противном случае, текущий
        if not isinstance(item, str):
            selected = item.selector.currentText()
        else:
            selected = self.current_chat

        try:
            self.transport.remove_contact(selected)
        except ServerError as err:
            self.messages.critical(self, 'Ошибка сервера', err.text)
        except OSError as err:
            if err.errno:
                self.messages.critical(
                    self, 'Ошибка', 'Потеряно соединение с сервером!')
                self.close()
            self.messages.critical(self, 'Ошибка', 'Таймаут соединения!')
        else:
            self.database.del_contact(selected)
            self.clients_list_update()
            logger.info(f'Успешно удалён контакт {selected}')
            if not isinstance(item, str):
                item.close()
            # Если удалён активный пользователь, то деактивируем поля ввода.
            if selected == self.current_chat:
                self.current_chat = None
                self.set_disabled_input()

    def send_message(self):
        """ Функция отправки сообщения пользователю. """
        # Текст в поле, проверяем что поле не пустое затем забирается сообщение
        # и поле очищается
        message_text = self.ui.text_message.toPlainText()
        self.ui.text_message.clear()
        if not message_text:
            return
        try:
            self.transport.send_message(self.current_chat, message_text)
            pass
        except ServerError as err:
            self.messages.critical(self, 'Ошибка', err.text)
        except OSError as err:
            if err.errno:
                self.messages.critical(
                    self, 'Ошибка', 'Потеряно соединение с сервером!')
                self.close()
            self.messages.critical(self, 'Ошибка', 'Таймаут соединения!')
        except (ConnectionResetError, ConnectionAbortedError):
            self.messages.critical(
                self, 'Ошибка', 'Потеряно соединение с сервером!')
            self.close()
        else:
            self.database.save_message(
                self.transport.username,
                self.current_chat,
                message_text)
            logger.debug(
                f'Отправлено сообщение для {self.current_chat}: '
                f'{message_text}')
            self.history_list_update()

    @pyqtSlot(str)
    def message(self, sender):
        """ Слот приёма нового сообщений """
        if sender == self.current_chat:
            self.history_list_update()
        else:
            # Проверим есть ли такой пользователь у нас в контактах:
            if self.database.check_contact(sender):
                # TODO Реализовать выделение чата
                if True:
                    self.current_chat = sender
                    self.set_active_user()
            else:
                # TODO Реализовать выделение нового чата
                if True:
                    self.add_contact(sender)
                    self.current_chat = sender
                    self.set_active_user()

    @pyqtSlot()
    def connection_lost(self):
        """ Слот потери соединения. Выдаёт сообщение об ошибке, и завершает
        работу приложения"""
        self.messages.warning(
            self,
            'Сбой соединения',
            'Потеряно соединение с сервером. ')
        self.close()

    def make_connection(self, trans_obj):
        """Соединение с сервером"""
        trans_obj.new_message.connect(self.message)
        trans_obj.connection_lost.connect(self.connection_lost)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'client.ui'
#
# Created by: PyQt5 UI code generator 5.15.4
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.

from PyQt5 import QtCore, QtWidgets


class UiMainClientWindow(object):
    """ Класс основного окна клиента. """
    def __init__(self):
        self.label_contacts = None
        self.central_widget = None
        self.btn_add_contact = None
        self.btn_remove_contact = None
        self.label_history = None
        self.text_message = None
        self.label_new_message = None
        self.list_contacts = None
        self.list_messages = None
        self.btn_send = None
        self.btn_clear = None
        self.menubar = None
        self.menu = None
        self.menu_2 = None
        self.statusBar = None
        self.menu_exit = None
        self.menu_add_contact = None
        self.menu_del_contact = None

    def setup_ui(self, main_client_window):
        main_client_window.setObjectName('main_client_window')
        main_client_window.resize(852, 565)
        main_client_window.setMinimumSize(QtCore.QSize(756, 534))
        self.central_widget = QtWidgets.QWidget(main_client_window)
        self.central_widget.setObjectName('central_widget')
        self.label_contacts = QtWidgets.QLabel(self.central_widget)
        self.label_contacts.setGeometry(QtCore.QRect(160, 10, 101, 16))
        self.label_contacts.setObjectName('label_contacts')
        self.btn_add_contact = QtWidgets.QPushButton(self.central_widget)
        self.btn_add_contact.setGeometry(QtCore.QRect(20, 30, 131, 131))
        self.btn_add_contact.setObjectName('btn_add_contact')
        self.btn_remove_contact = QtWidgets.QPushButton(self.central_widget)
        self.btn_remove_contact.setGeometry(QtCore.QRect(20, 170, 131, 111))
        self.btn_remove_contact.setObjectName('btn_remove_contact')
        self.label_history = QtWidgets.QLabel(self.central_widget)
        self.label_history.setGeometry(QtCore.QRect(390, 10, 431, 21))
        self.label_history.setObjectName('label_history')
        self.text_message = QtWidgets.QTextEdit(self.central_widget)
        self.text_message.setGeometry(QtCore.QRect(390, 430, 441, 91))
        self.text_message.setObjectName('text_message')
        self.label_new_message = QtWidgets.QLabel(self.central_widget)
        self.label_new_message.setGeometry(QtCore.QRect(390, 410, 441, 16))
        self.label_new_message.setObjectName('label_new_message')
        self.list_contacts = QtWidgets.QListView(self.central_widget)
        self.list_contacts.setGeometry(QtCore.QRect(160, 30, 211, 491))
        self.list_contacts.setObjectName('list_contacts')
        self.list_messages = QtWidgets.QListView(self.central_widget)
        self.list_messages.setGeometry(QtCore.QRect(390, 30, 441, 371))
        self.list_messages.setObjectName('list_messages')
        self.btn_send = QtWidgets.QPushButton(self.central_widget)
        self.btn_send.setGeometry(QtCore.QRect(20, 410, 131, 111))
        self.btn_send.setObjectName('btn_send')
        self.btn_clear = QtWidgets.QPushButton(self.central_widget)
        self.btn_clear.setGeometry(QtCore.QRect(20, 290, 131, 111))
        self.btn_clear.setObjectName('btn_clear')
        main_client_window.setCentralWidget(self.central_widget)
        self.menubar = QtWidgets.QMenuBar(main_client_window)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 852, 22))
        self.menubar.setObjectName('menubar')
        self.menu = QtWidgets.QMenu(self.menubar)
        self.menu.setObjectName('menu')
        self.menu_2 = QtWidgets.QMenu(self.menubar)
        self.menu_2.setObjectName('menu_2')
        main_client_window.setMenuBar(self.menubar)
        self.statusBar = QtWidgets.QStatusBar(main_client_window)
        self.statusBar.setObjectName('statusBar')
        main_client_window.setStatusBar(self.statusBar)
        self.menu_exit = QtWidgets.QAction(main_client_window)
        self.menu_exit.setObjectName('menu_exit')
        self.menu_add_contact = QtWidgets.QAction(main_client_window)
        self.menu_add_contact.setObjectName('menu_add_contact')
        self.menu_del_contact = QtWidgets.QAction(main_client_window)
        self.menu_del_contact.setObjectName('menu_del_contact')
        self.menu.addAction(self.menu_exit)
        self.menu_2.addAction(self.menu_add_contact)
        self.menu_2.addAction(self.menu_del_contact)
        self.menu_2.addSeparator()
        self.menubar.addAction(self.menu.menuAction())
        self.menubar.addAction(self.menu_2.menuAction())

        self.retranslate_ui(main_client_window)
        self.btn_clear.clicked.connect(self.text_message.clear)
        QtCore.QMetaObject.connectSlotsByName(main_client_window)

    def retranslate_ui(self, main_client_window):
        _translate = QtCore.QCoreApplication.translate
        main_client_window.setWindowTitle(
            _translate('main_client_window', 'Чат программа alpha release'))
        self.label_contacts.setText(_translate('main_client_window',
                                               'Контакты'))
        self.btn_add_contact.setText(_translate('main_client_window',
                                                'Добавить'))
        self.btn_remove_contact.setText(_translate('main_client_window',
                                                   'Удалить контакт'))
        self.label_history.setText(_translate('main_client_window', 'История'))
        self.label_new_message.setText(_translate('main_client_window',
                                                  'Отправить новое сообщение'))
        self.btn_send.setText(_translate('main_client_window', 'Отправить'))
        self.btn_clear.setText(_translate('main_client_window', 'Очистить'))
        self.menu.setTitle(_translate('main_client_window', 'Файл'))
        self.menu_2.setTitle(_translate('main_client_window', 'Контакты'))
        self.menu_exit.setText(_translate('main_client_window', 'Выход'))
        self.menu_add_contact.setText(_translate('main_client_window',
                                                 'Добавить контакт'))
        self.menu_del_contact.setText(_translate('main_client_window',
                                                 'Удалить контакт'))
